from redpepper.operations import Operation, Result


class Show(Operation):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"data.Show({self.name})"

    def run(self, agent):
        result = Result(self)
        data = agent.request("data", name=self.name)
        result.succeeded = True
        result += data
        return result



from redpepper.operations import Operation, Result


class Echo(Operation):
    def __init__(self, message, reverse=False):
        self.message = message
        self.reverse = reverse

    def __str__(self):
        return f'echo.Echo("{self.message}"{" reverse" if self.reverse else ""})'

    def run(self, agent):
        result = Result(self)
        message = self.message
        if self.reverse:
            message = message[::-1]
        result += message
        return result



import base64
import grp
import hashlib
import io
import logging
import os
import pwd

from redpepper.operations import Operation, Result

logger = logging.getLogger(__name__)


class Installed(Operation):
    def __init__(
        self,
        path: str,
        source: str,
        method: str = "hash",
        overwrite: bool = True,
        user: str | int | None = None,
        group: str | int | None = None,
        mode: int | None = None,
    ):
        self.path = path
        self.source = source
        self.method = method
        if method not in ["stat", "hash", "content"]:
            raise ValueError(f"Unknown method: {method}")
        self.overwrite = overwrite
        if user is not None and isinstance(user, str):
            self.user = pwd.getpwnam(user).pw_uid
        else:
            self.user = user
        if group is not None and isinstance(group, str):
            self.group = grp.getgrnam(group).gr_gid
        else:
            self.group = group
        if mode is not None and not isinstance(mode, int):
            mode = int(mode, 8)
        self.mode = mode

    def __str__(self):
        if self.method == "content":
            source = "content"
        else:
            source = f'"{self.source}"'
        return f'file.Installed("{self.path}" from {source})'

    def ensure(self, agent):
        result = Result(self)
        # Open in binary writable mode without truncating the file
        try:
            f = open(self.path, "r+b")
            logger.debug("File %r already exists", self.path)
        except FileNotFoundError:
            f = open(self.path, "wb")
            logger.debug("Creating file %r", self.path)
        mtime = None
        with f:
            stat = os.fstat(f.fileno())
            if self.mode is not None and stat.st_mode & 0o777 != self.mode:
                os.fchmod(f.fileno(), self.mode)
                result += (
                    f"Changed mode from 0{stat.st_mode & 0o777:o} to 0{self.mode:o}."
                )
                result.changed = True
            if self.user is not None and stat.st_uid != self.user:
                os.fchown(f.fileno(), self.user, -1)
                result += f"Changed owner from {stat.st_uid} to {self.user}."
                result.changed = True
            if self.group is not None and stat.st_gid != self.group:
                os.fchown(f.fileno(), -1, self.group)
                result += f"Changed group from {stat.st_gid} to {self.group}."
                result.changed = True
            if self.overwrite:
                try:
                    nwritten, mtime = self.ensure_file_contents(agent, f)
                except Exception:
                    result.exception()
                    return result
                else:
                    if nwritten is not None:
                        result += f"Wrote {nwritten} bytes to {self.path}."
                        result.changed = True
        if mtime is not None:
            os.utime(self.path, (mtime, mtime))
        if not result.changed:
            result += f'File "{self.path}" is already in the specified state.'
        return result

    def ensure_file_contents(self, agent, f: io.BufferedIOBase):
        logger.debug("Comparing file using %s method", self.method)
        rewrite = False
        if self.method == "content":
            try:
                existing = f.read()
            except io.UnsupportedOperation:
                existing = b""
            shouldbe = self.source.encode("utf-8")
            if existing != shouldbe:
                f.seek(0)
                f.write(shouldbe)
                f.truncate()
                f.flush()
                os.fsync(f.fileno())
                nwritten = len(shouldbe)
                return nwritten, None
            return None, None
        remote_stat = agent.request("dataFileStat", path=self.source)
        if self.method == "stat":
            try:
                existing_stat = os.fstat(f.fileno())
                existing_mtime = existing_stat.st_mtime
                existing_size = existing_stat.st_size
            except FileNotFoundError:
                existing_mtime = None
                existing_size = None
            logger.debug(
                "Mtime of %r: %s vs. %s",
                self.path,
                existing_mtime,
                remote_stat["mtime"],
            )
            logger.debug(
                "Size of %r: %s vs. %s", self.path, existing_size, remote_stat["size"]
            )
            rewrite = (
                existing_mtime != remote_stat["mtime"]
                or existing_size != remote_stat["size"]
            )
        elif self.method == "hash":
            hash = agent.request("dataFileHash", path=self.source)
            existing_hash = self.hash_file(f)
            logger.debug("Hash of %s: %s vs. %s", self.path, existing_hash, hash)
            rewrite = existing_hash != hash
        if not rewrite:
            return None, None
        # Retrieve the entire file into the buffer first so that we don't
        # leave the file in an inconsistent state if the connection is lost
        contents = io.BytesIO()
        while True:
            data = agent.request(
                "dataFileContents",
                filename=self.source,
                offset=contents.tell(),
                length=32 * 1024,
            )
            data = base64.b64decode(data)
            if not data:
                break
            contents.write(data)
        f.seek(0)
        f.write(contents.getbuffer())
        f.truncate()
        f.flush()
        os.fsync(f.fileno())
        nwritten = contents.tell()
        contents.close()
        return nwritten, remote_stat["mtime"]

    def hash_file(self, f: io.BufferedIOBase):
        try:
            hash = hashlib.sha256()
            for chunk in iter(lambda: f.read(4096), b""):
                hash.update(chunk)
            return hash.hexdigest()
        except io.UnsupportedOperation:
            return None


class Symlinked(Operation):
    _no_changes_text = "The symlink is already in the specified state."

    def __init__(self, path, target):
        self.path = path
        self.target = target

    def __str__(self):
        return f'file.Symlinked("{self.path}" to "{self.target}")'

    def test(self, agent):
        try:
            existing_target = os.readlink(self.path)
        except FileNotFoundError:
            existing_target = None
        return existing_target == self.target

    def run(self, agent):
        result = Result(self)
        try:
            existing_target = os.readlink(self.path)
        except FileNotFoundError:
            existing_target = None
        if existing_target != self.target:
            try:
                os.symlink(self.target, self.path)
            except FileExistsError:
                os.remove(self.path)
                os.symlink(self.target, self.path)
            result += f"Symlinked {self.path} to {self.target}."
            result.changed = True
        return result


class Dir(Operation):
    def __init__(self, path, user=None, group=None, mode=None):
        self.path = path
        self.user = user
        if user is not None and not isinstance(user, int):
            self.user = pwd.getpwnam(user).pw_uid
        self.group = group
        if group is not None and not isinstance(group, int):
            self.group = grp.getgrnam(group).gr_gid
        if mode is not None and not isinstance(mode, int):
            mode = int(mode, 8)
        self.mode = mode

    def __str__(self):
        return f'file.DirExists("{self.path}")'

    def test(self, agent):
        if not os.path.isdir(self.path):
            return False
        stat = os.stat(self.path)
        if self.mode is not None and stat.st_mode & 0o777 != self.mode:
            return False
        if self.user is not None and stat.st_uid != self.user:
            return False
        if self.group is not None and stat.st_gid != self.group:
            return False
        return True

    def run(self, agent):
        result = Result(self)
        try:
            stat = os.stat(self.path)
        except FileNotFoundError:
            os.mkdir(self.path)
            stat = os.stat(self.path)
            result += f"Created directory {self.path}."
            result.changed = True
        if self.mode is not None and stat.st_mode & 0o777 != self.mode:
            os.chmod(self.path, self.mode)
            result += f"Changed mode from 0{stat.st_mode & 0o777:o} to 0{self.mode:o}."
            result.changed = True
        if self.user is not None and stat.st_uid != self.user:
            os.chown(self.path, self.user, -1)
            result += f"Changed owner from {stat.st_uid} to {self.user}."
            result.changed = True
        if self.group is not None and stat.st_gid != self.group:
            os.chown(self.path, -1, self.group)
            result += f"Changed group from {stat.st_gid} to {self.group}."
            result.changed = True
        return result



import sys

if sys.platform == "linux":
    from redpepper.operations.systemd import Enabled, Reload, Restart, Running  # noqa
else:
    raise ImportError(f"Unsupported platform for service module: {sys.platform}")



import os
import pwd
import subprocess

from redpepper.operations import Operation, Result


class UpToDate(Operation):
    def __init__(
        self,
        target,
        repo,
        force_pull=False,
        rebase=None,
        ff_only=False,
        user=None,
        identity=None,
        depth=None,
        recurse_submodules=False,
    ):
        self.target = target
        self.repo = repo
        self.force_pull = force_pull
        self.rebase = rebase
        self.ff_only = ff_only
        self.user = user
        self.identity = identity
        self.depth = depth
        self.recurse_submodules = recurse_submodules

    def __str__(self):
        return f'git.UpToDate({self.target} from "{self.repo}")'

    def ensure(self, agent):
        result = Result(self)
        clone = False
        if not os.path.isdir(self.target):
            clone = True
        elif not os.path.isdir(os.path.join(self.target, ".git")):
            # git allows empty directories
            clone = True

        git_cmd = ["git"]
        if clone:
            git_cmd.append("clone")
            git_cmd.append(self.repo)
            git_cmd.append(self.target)
        else:
            git_cmd.append("-C")
            git_cmd.append(self.target)
            git_cmd.append("pull")
            if self.force_pull:
                git_cmd.append("--force")
            if self.rebase is not None:
                if self.rebase:
                    git_cmd.append("--rebase")
                else:
                    git_cmd.append("--no-rebase")
            if self.ff_only:
                git_cmd.append("--ff-only")
        if self.depth:
            git_cmd.append("--depth")
            git_cmd.append(str(self.depth))
        if self.recurse_submodules:
            git_cmd.append("--recurse-submodules")

        kw = {}
        kw["env"] = os.environ.copy()
        if self.user:
            kw["user"] = self.user
            kw["env"]["HOME"] = pwd.getpwnam(self.user).pw_dir
        if self.identity:
            kw["env"]["GIT_SSH_COMMAND"] = f'ssh -i "{self.identity}"'

        p = subprocess.run(
            git_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, **kw
        )
        if not result.check_completed_process(p).succeeded:
            return result

        if clone:
            result.changed = True
        elif "Already up to date" not in p.stdout:
            result.changed = True

        return result



import logging
import subprocess

from redpepper.operations import Operation, Result

logger = logging.getLogger(__name__)


class Installed(Operation):
    _no_changes_text = "The package is already installed."

    def __init__(self, name, env={}):
        self.name = name
        if not isinstance(env, dict):
            raise ValueError("env must be a mapping")
        self.env = env

    def __str__(self):
        return f"apt.Installed({self.name})"

    def test(self, agent):
        cmd = [
            "dpkg-query",
            "--showformat",
            "${Package}\t${db:Status-Status}\n",
            "-W",
            self.name,
        ]
        logger.debug(
            "Testing if apt package %s is installed with command %r", self.name, cmd
        )
        p = subprocess.run(cmd, capture_output=True, text=True)
        if p.returncode == 1:
            logger.debug("dpkg-query returned 1")
            return False
        elif p.returncode == 0:
            logger.debug("dpkg-query returned 0, output: %r", p.stdout)
            for line in p.stdout.splitlines():
                package, status = line.split("\t")
                if package != self.name:
                    continue
                return status == "installed"
            return False
        else:
            logger.error("dpkg-query failed with return code %s", p.returncode)
            raise subprocess.CalledProcessError(p.returncode, cmd, p.stdout, p.stderr)

    def run(self, agent):
        result = Result(self)
        cmd = [
            "apt-get",
            "-q",  # quiet
            "-y",  # assume yes
            "-o",
            "DPkg::Options::=--force-confold",  # keep old config files
            "-o",
            "DPkg::Options::=--force-confdef",  # keep new config files
            "install",  # install package
            self.name,  # package name
        ]
        env = self.env
        env["DEBIAN_FRONTEND"] = "noninteractive"  # don't ask questions
        env["APT_LISTCHANGES_FRONTEND"] = "none"  # don't show changelogs
        env["APT_LISTBUGS_FRONTEND"] = "none"  # don't show bug reports
        env["UCF_FORCE_CONFFOLD"] = "1"  # keep old config files
        p = subprocess.run(cmd, capture_output=True, text=True, env=env)
        if result.check_completed_process(p).succeeded:
            result.changed = "Setting up" in p.stdout
        return result


class UnattendedUpgrade(Operation):
    def __str__(self):
        return "apt.UnattendedUpgrade()"

    def run(self, agent):
        result = Result(self)
        if not result.update(Installed("unattended-upgrades").ensure(agent)).succeeded:
            return result
        p = subprocess.run(["unattended-upgrades"], capture_output=True, text=True)
        if result.check_completed_process(p).succeeded:
            result.changed = result.changed or p.stdout != ""
        return result



import importlib.util
import os

_module_hook = None


def __getattr__(name, use_hook=True):
    from . import custom_operations_dir

    if use_hook and _module_hook:
        _module_hook(name)

    file = os.path.join(custom_operations_dir, name + ".py")
    if os.path.exists(file):
        try:
            spec = importlib.util.spec_from_file_location(name, file)
            if spec is None or spec.loader is None:
                raise ImportError(f"Failed to load operation module {name}")
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
        except Exception as e:
            raise ImportError(f"Failed to load operation module {name}") from e
        return module
    raise AttributeError(f"module {__name__} has no attribute {name}")



import os
import pwd
import subprocess

from redpepper.operations import Operation, Result


class Run(Operation):
    def __init__(
        self,
        command,
        user=None,
        group=None,
        shell=True,
        cwd=None,
        stdin=None,
        wait=True,
        env=None,
        capture_stdout=True,
        capture_stderr=True,
        encoding="utf-8",
    ):
        self.command = command
        self.user = user
        self.group = group
        self.shell = shell
        self.cwd = cwd
        if stdin is not None:
            self.stdin = stdin.encode(encoding)
        else:
            self.stdin = None
        self.wait = wait
        if env:
            if not isinstance(env, dict):
                raise TypeError("env must be a mapping of environment variables")
            self.env = {}
            for k, v in env.items():
                if isinstance(v, (str, int, float, bool)):
                    self.env[k] = str(v)
                else:
                    raise TypeError(
                        f"Environment variable {k} has invalid type {type(v).__name__}"
                    )
        else:
            self.env = None
        self.capture_stdout = capture_stdout
        self.capture_stderr = capture_stderr
        self.encoding = encoding
        if not wait and (capture_stdout or capture_stderr):
            raise ValueError(
                "Cannot capture output if not waiting for command to finish"
            )

    def __str__(self):
        return f'Run("{self.command}{"" if self.wait else " &"}"{" as " + self.user if self.user else ""})'

    def run(self, agent):
        result = Result(self)
        kw = {}
        kw["env"] = os.environ.copy()
        if self.user:
            kw["user"] = self.user
            kw["env"]["HOME"] = pwd.getpwnam(self.user).pw_dir
        if self.group:
            kw["group"] = self.group
        elif self.user:
            kw["group"] = pwd.getpwnam(self.user).pw_gid
        if self.shell:
            kw["shell"] = True
        if self.cwd:
            kw["cwd"] = self.cwd
        if self.env is not None:
            kw["env"].update(self.env)
        if self.stdin is not None:
            kw["stdin"] = subprocess.PIPE
        if self.capture_stdout:
            kw["stdout"] = subprocess.PIPE
        if self.capture_stderr:
            kw["stderr"] = subprocess.PIPE
        process: subprocess.Popen[bytes] = subprocess.Popen(self.command, **kw)  # type: ignore
        if not self.wait:
            result += f"$ {self.command} &"
            return result
        stdout, stderr = process.communicate(self.stdin)
        if self.capture_stdout and stdout:
            result += stdout.decode(self.encoding, errors="replace").rstrip()
        if self.capture_stderr and stderr:
            result += (
                "Stderr:\n" + stderr.decode(self.encoding, errors="replace").rstrip()
            )
        if process.returncode != 0:
            result.fail(f"Command failed with return code {process.returncode}")
        result.changed = True
        return result


class RunMultiple(Operation):
    def __init__(self, commands, **kw):
        self.commands = [Run(c, **kw) for c in commands]
        self.kw = kw

    def __str__(self):
        return f"RunMultiple({len(self.commands)} commands)"

    def run(self, agent):
        result = Result(self)
        for cmd in self.commands:
            result += f"Running {cmd}:"
            if not result.update(cmd.run(agent)).succeeded:
                return result
        return result



import os
import subprocess
import sys

from redpepper.operations import Operation, Result

if sys.platform != "linux":
    raise ImportError("sysctl module only supported on Linux")

SYSCTL_VERSION = int(
    subprocess.run(
        ["dpkg-query", "--showformat", "${Version}\n", "-W", "systemd"],
        capture_output=True,
        text=True,
        check=True,
    )
    .stdout.strip()
    .split(".")[0]
)

# There is already a 99-sysctl.conf, so make this start with "99-x"
SYSCTL_CONF_PATH = "/etc/sysctl.d/99-x-redpepper.conf"
if SYSCTL_VERSION < 207:
    SYSCTL_CONF_PATH = "/etc/sysctl.conf"


class Parameter(Operation):
    def __init__(self, name, value):
        self.name = name
        self.value = str(value)

    def __str__(self):
        return f"sysctl.Parameter({self.name} = {self.value})"

    def test(self, agent):
        if not os.path.exists(SYSCTL_CONF_PATH):
            return False
        with open(SYSCTL_CONF_PATH, "r") as f:
            for line in f:
                line = line.strip()
                if line.startswith("#") or "=" not in line:
                    continue
                k, v = line.split("=")
                if k.strip() == self.name and v.strip() == self.value:
                    return True
        return False

    def run(self, agent):
        result = Result(self)
        with open(SYSCTL_CONF_PATH, "a") as f:
            f.write(f"\n{self.name} = {self.value}\n")
        result += (
            f'Added sysctl parameter "{self.name} = {self.value}" to {SYSCTL_CONF_PATH}'
        )
        result.changed = True
        p = subprocess.run(["sysctl", "--system"], capture_output=True, text=True)
        result.check_completed_process(p)
        return result
